(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

open Parsetree;;

(* test functions on the type relation *)

let is_rewrite = function Rewrite _ -> true | _ -> false;;
let is_commutative = function Commutative _ -> true | _ -> false;;
let is_associative = function Associative _ -> true | _ -> false;;
let is_absorbent = function Absorbent _ -> true | _ -> false;;
let is_absorbing = function Absorbing _ -> true | _ -> false;;
let is_neutral = function Neutral _ -> true | _ -> false;;
let is_idempotent = function Idempotent _ -> true | _ -> false;;
let is_inverse = function Inverse _ -> true | _ -> false;;
let is_distributive = function Distributive _ -> true | _ -> false;;

(* test functions on lists of relations *)

let has_commutative = List.exists is_commutative;;
let has_associative = List.exists is_associative;;
let has_idempotent = List.exists is_idempotent;;
let has_absorbing = List.exists is_absorbing;;

(* access function on lists of relations *)

let rec rewrite = function
  | Rewrite x :: l -> x :: rewrite l
  | _ :: l -> rewrite l
  | [] -> [];;

let rec commutative = function
  | Commutative s :: _ -> Some s
  | _ :: l -> commutative l
  | _ -> None;;

let rec associative = function
  | Associative s :: _ -> Some s
  | _ :: l -> associative l
  | _ -> None;;

let rec absorbent = function
  | Absorbent x :: _ -> Some x
  | _ :: l -> absorbent l
  | [] -> None;;

let rec absorbing = function
  | Absorbing x :: _ -> Some x
  | _ :: l -> absorbing l
  | [] -> None;;

let rec neutral = function
  | Neutral x :: _ -> Some x
  | _ :: l -> neutral l
  | [] -> None;;

let rec idempotent = function
  | Idempotent s :: _ -> Some s
  | _ :: l -> idempotent l
  | _ -> None;;

let rec inverse = function
  | Inverse x :: _ -> Some x
  | _ :: l -> inverse l
  | [] -> None;;

let rec distributive = function
  | Distributive x :: l -> x :: distributive l
  | _ :: l -> distributive l
  | [] -> [];;

(* string of relation *)

let string_of_side = function
  | Right -> "Right"
  | Left -> "Left"
  | Both -> "Both";;

let name_of_relation = function
  | Rewrite _ -> "Rule"
  | Commutative s -> "Commutative" ^ string_of_side s
  | Associative s -> "Associative" ^ string_of_side s
  | Absorbent (s, _) -> "Absorbent" ^ string_of_side s
  | Absorbing (s, _) -> "Absorbing" ^ string_of_side s
  | Neutral (s, _) -> "Neutral" ^ string_of_side s
  | Idempotent s -> "Idempotent" ^ string_of_side s
  | Idempotent_Unary -> "Idempotent_Unary"
  | Idempotent_AC -> "Idempotent_AC"
  | Nilpotent _ -> "Nilpotent"
  | Involutive -> "Involutive"
  | Inverse (s, _, _) -> "Inverse" ^ string_of_side s
  | Distributive (s, _) -> "Distributive" ^ string_of_side s
  | Distributive_Inverse (s, _) -> "Distributive_Inverse" ^ string_of_side s
  | Distributive_Unary (_, _) -> "Distributive_Unary"
  | Inverse_Neutral (_, _) -> "Inverse_Neutral"
  | Division_by_Absorbent _ -> "Division_by_Absorbent";;

(* Computing precedences from an ordered list of the properties with
   dummy arguments. *)

let underscore_pat = { ppat_desc = Ppat_any; ppat_loc = Location.none };;

let precedences =
 let t = Hashtbl.create 37 in
 let p = ref 0 in
 List.iter
  (fun rel -> incr p; Hashtbl.add t (name_of_relation rel) !p)
  [Division_by_Absorbent "";
   Rewrite (underscore_pat, None, underscore_pat);
   Nilpotent "";
   Absorbent (Right, "");
   Absorbent (Left, "");
   Absorbent (Both, "");
   Neutral (Right, "");
   Neutral (Left, "");
   Neutral (Both, "");
   Inverse_Neutral ("", None);
   Involutive;
   Distributive_Unary ("", None);
   Distributive (Right, "");
   Distributive (Left, "");
   Distributive (Both, "");
   Distributive_Inverse (Right, "");
   Distributive_Inverse (Left, "");
   Distributive_Inverse (Both, "");
   Associative Right;
   Associative Left;
   Associative Both;
   Idempotent_Unary;
   Absorbing (Right, "");
   Absorbing (Left, "");
   Absorbing (Both, "");
   Inverse (Right, "", None);
   Inverse (Left, "", None);
   Inverse (Both, "", None);
   Commutative Right;
   Commutative Left;
   Commutative Both;
   Idempotent_AC;
   Idempotent Right;
   Idempotent Left;
   Idempotent Both;
  ];
 t;;

let precedence_of_relation r = Hashtbl.find precedences (name_of_relation r);;

let compare_relation r1 r2 =
  compare (precedence_of_relation r1) (precedence_of_relation r2);;

module RelOrd = struct
  type t = relation;;
  let compare = compare_relation;;
end;;

module RelSet = Set.Make (RelOrd);;
