(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: complete.ml,v 1.11 2007/04/25 15:04:50 weis Exp $ *)

open Spec;;
open Relation;;
open Parsetree;;

open Genr_base;;

(* Builds a function definition from a constructor definition. *)

let genr_fdef cdef = {
    fd_uname = cdef.cd_name;
    fd_relations = List.sort compare_relation cdef.cd_relations;
    fd_arity = cdef.cd_arity;
  };;

(* Creates a hash table [constructor name -> function definition]
   from a list of constructor definitions. *)

let create_table cdefs =
  let ht = Hashtbl.create 17 in
  List.iter (fun cdef -> Hashtbl.add ht cdef.cd_name (genr_fdef cdef)) cdefs;
  ht;;

(* Add properties [xs] in a list of properties [l], only if they are
   not already member of [l].
   The resulting list is sorted wrt the predicate [compare_relation]. *)

(* let add_uniq x l = if List.mem x l then l else x :: l;;

let add_props xs l = List.sort compare_relation (List.fold_right add_uniq xs l);; *)

(* optimisation: we assume that the list argument is already sorted *)

let add_uniq x =
  let rec add = function
    | [] -> [x]
    | (y :: _) as l when compare_relation x y < 0 -> x :: l
    | (y :: _) as l when x = y -> l
    | y :: l -> y :: add l
  in add;;

let add_props = List.fold_right add_uniq;;

(* Function for adding properties to [dname] in a hash table
   [constructor name -> function definition]. *)

let update_props ht dname new_props =
  let fdef = Hashtbl.find ht dname in
  let fdef = { fdef with fd_relations = add_props new_props fdef.fd_relations } in
  Hashtbl.replace ht dname fdef;;

(* Adds in [ht] the properties required by [ap]. *)

let update_by_ap ht fdef =

  let cname = fdef.fd_uname in

  let add_division_by_absorbent e props =
    match absorbent fdef.fd_relations with
    | None -> props
    | Some (_side, abs) ->
      (* In some cases, the neutral element is equal to the absorbent element,
         hence we have not to generate Division_by_Absorbent. *)
      if abs <> e then Division_by_Absorbent abs :: props else props in

  let e_opt =
    match neutral fdef.fd_relations with
    | Some (_side, e) -> Some e
    | None -> None in

  let rec update = function
  | Inverse (s, inv, None) ->
    (* A true inverse we need a neutral element. *)
    let e =
      match e_opt with
      | Some e -> e
      | None ->
        failwith "No neutral element for implicit inverse operation."
        (* Need a neutral element for s *) in
    (* Properties of inverse. *)
    let props =
      [Involutive;
       Distributive_Unary (cname, Some cname)] in
    (* If the operator c is commutative then
         rinv re = re (for right inverse)
         linv le = le (for left inverse)
         inv e = e (for inverse Both). *)
    let props =
      if has_commutative fdef.fd_relations
      then Inverse_Neutral (e, Some e) :: props
      else props in
    let props = add_division_by_absorbent e props in
    update_props ht inv props

  | Inverse (side, inv, Some a) ->
    (* If a is the neutral element, then inv is a true inverse,
       hence we just have to call update with the proper argument. *)
    if Some a = e_opt then update (Inverse (side, inv, None)) else
    (* Inv is a fake inverse: we treat it specially. *)
    let props =
      [Involutive; ] in
    let props =
      match e_opt with
      | Some e ->
        Inverse_Neutral (e, Some a) ::
        Inverse_Neutral (a, Some e) :: props
      | None -> props in
    let props = add_division_by_absorbent a props in
    update_props ht inv props

  | _ -> () in

  update;;

(* Adds in [ht] the properties required by the properties of [cdefs]. *)

let update_table ht cdefs =
  let update cname fdef =
    List.iter (update_by_ap ht fdef) fdef.fd_relations in
  Hashtbl.iter update ht;;

(* Gives the list of function definitions corresponding to a list of
   constructor definitions, after having added all the required
   properties. *)
let genr_fdefs cdefs =
  let ht = create_table cdefs in
  update_table ht cdefs;
  Hashtbl.fold (fun _ fdef l -> fdef :: l) ht [];;
