(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr_binary.ml,v 1.25 2007/04/26 17:07:58 blanqui Exp $ *)

open Parsetree;;
open Relation;;
open Code;;

open Genr_base;;

let insert_in = prefix_cname "insert";;
let insert_inv_in = prefix_cname "insert_inv";;
let delete_in = prefix_cname "delete";;
let is_redex_in = prefix_cname "is_redex";;
let return_in = prefix_cname "return";;

(* Application of the relevant generated function for the constructor cname:
   the relevant app* is used, the proper name is generated and arity is taken
   into account. *)
let insert cname = app2 (insert_in cname);;
let insert_inv cname = app2 (insert_inv_in cname);;
let delete cname = app2 (delete_in cname);;
let is_redex cname x y = app1 (is_redex_in cname) (pair x y);;
let return cname x y = app1 (return_in cname) (pair x y);;

(* Functions for selecting clauses *)

let clauses_of_side s left right =
  match s with
  | Left -> [left]
  | Right -> [right]
  | Both -> [left; right];;

let clauses_left_or_right_of_side s left right =
  match s with
  | Left | Both -> [left]
  | Right -> [right];;

let opt b c = if b then [c] else [];;

(* [insert_right x u] inserts x in the right comb u *)

let genr_insert fdef =
  let cname = fdef.fd_uname in

  (* arguments of insert *)
  let args = Code.idents_of_exps [x; u] in

  let idem = has_idempotent fdef.fd_relations
  and commut = has_commutative fdef.fd_relations in

  (* we match on u *)
  let clauses =
    opt commut (Clause_when (pattern2 cname y underscore, appi x "<" y, return cname x u))
    @ opt idem (Clause_when (pattern2 cname y underscore, appi x "=" y, u))
    @ opt commut (Clause (pattern2 cname y t, return cname y (insert cname x t)))
    @ opt commut (Clause_when (underscore, appi x ">" u, return cname u x))
    @ opt (idem && commut) (Clause_when (underscore, appi x "=" u, u))
    @ [Clause (underscore, return cname x u)]
  in insert_in cname, args, Match (u, clauses);;

(* [delete x u] tries do delete x in the right comb u, raise Not_found otherwise *)

let genr_delete cname ename =

  (* arguments of delete *)
  let args = Code.idents_of_exps [x; u] in

  (* we match on u *)
  let clauses = [
    Clause_when (pattern2 cname y underscore, appi x "<" y, raise_error "Not_found");
    Clause_when (pattern2 cname y t, appi x "=" y, constr_fun2 cname (Id ename) t);
    Clause (pattern2 cname y t, constr2 cname y (delete cname x t));
    Clause_when (underscore, appi u "=" x, Id ename);
    Clause (underscore, raise_error "Not_found");
  ] in
    delete_in cname, args, Match (u, clauses);;

(* [insert_inv x u] tries do delete x in the right comb u, otherwise inserts (opp x) in u *)

let genr_insert_inv cname oname =
  let opp = app1 (construction_function_name oname) in
 
  (* arguments of insert_inv *)
  let args = Code.idents_of_exps [x; u] in

  let clauses = [Clause (Id "Not_found", insert cname (opp x) u)] in
  let body = Try (delete cname x u, clauses) in
    insert_inv_in cname, args, body;;
    
(* Generate return function *)

let genr_return cname =

  (* arguments of return *)
  let args = Code.idents_of_exps [z] in
  let def = If (is_redex cname x y, constr_fun2 cname x y, constr2 cname x y) in
  let clauses = [Clause (pair x y, def)] in
    return_in cname, args, Match (z, clauses);;

(* Generating the relevant clauses of the construction function for
   the constructor given by [fdef], using the relations stated by the
   algebraic property [ap]. *)

let genr_fun_2_by_ap fdef =
  let cname = fdef.fd_uname in function

  | Rewrite (p1, eopt, p2) -> [genr_rule p1 eopt p2]

  | Commutative _s -> [] (* done in insert *)

  | Associative s ->
    let left_clause =
      Clause (pair x (pattern2 cname y z),
	     constr_fun2 cname (constr_fun2 cname x y) z)
    and right_clause =
      Clause (pair (pattern2 cname x y) z,
	     constr_fun2 cname x (constr_fun2 cname y z)) in
    clauses_left_or_right_of_side s left_clause right_clause

  | Absorbent (s, aname) ->
    (* c (a, _) -> a. *)
    let left_clause = Clause (pair (Id aname) underscore, Id aname)
    and right_clause = Clause (pair underscore (Id aname), Id aname) in
    clauses_of_side s left_clause right_clause

  | Neutral (s, ename) ->
    (* c (e, y) -> y. *)
    let left_clause = Clause (pair (Id ename) y, y)
    and right_clause = Clause (pair x (Id ename), x) in
    clauses_of_side s left_clause right_clause

  | Idempotent _s -> [] (* done in insert *)

  | Nilpotent _
  | Involutive -> assert false (* properties for unary constructors *)

  | Inverse (s, iname, _) ->
    (* c (inv(x), x) -> e. when e = neutral of c. *)
    let left_clause = Clause (pair (pattern1 iname x) y, insert_inv cname x y)
    and right_clause = Clause (pair x (pattern1 iname y), insert_inv cname y x) in
    clauses_of_side s left_clause right_clause

  | Distributive_Inverse (s, iname) ->
    (* c (inv (x), y) -> inv (c (x, y)). *)
    let left_clause =
      Clause (pair (pattern1 iname x) y, constr_fun1 iname (constr_fun2 cname x y))
    and right_clause =
      Clause (pair x (pattern1 iname y), constr_fun1 iname (constr_fun2 cname x y))
    in clauses_of_side s left_clause right_clause

  | Distributive (s, dname) ->
    (* c (d (x, y), z) -> d (c (x,z), c (y,z)). *)
    let left_clause =
      Clause (pair (pattern2 dname x y) z,
	      constr_fun2 dname (constr_fun2 cname x z) (constr_fun2 cname y z))
    and right_clause =
      Clause (pair x (pattern2 dname y z),
              constr_fun2 dname (constr_fun2 cname x y) (constr_fun2 cname x z)) in
    clauses_of_side s left_clause right_clause

  | Absorbing (s, dname) ->
    (* c (d (x, y), y) -> y. *)
    let left_clause = Clause_when (pair (pattern2 dname x y) z, appi y "=" z, y)
    and right_clause = Clause_when (pair x (pattern2 dname y z), appi x "=" y, x)
    in clauses_of_side s left_clause right_clause 

  | Idempotent_AC -> [] (* done in insert *)

  | Division_by_Absorbent _
  | Inverse_Neutral (_, _)
  | Distributive_Unary (_, _)
  | Idempotent_Unary -> assert false (* properties for unary constructors *)
  ;;

(* Generate a function checking whether a Rewrite rule can be applied *)

let genr_is_redex fdef =
  let cname = fdef.fd_uname in

  (* arguments of is_redex *)
  let args = Code.idents_of_exps [z] in

  (* let true_clause = function
  | Clause (l, _) -> Clause (l, Id "true")
  | Clause_when (l, c, _) -> Clause_when (l, c, Id "true") in
  let non_rewrite = function Rewrite _ -> false | _ -> true in
  let builtin_relations = List.filter non_rewrite fdef.fd_relations in
  let acc_clauses ap l = genr_fun_2_by_ap fdef ap @ l in
  let b_clauses = List.fold_right acc_clauses builtin_relations [] in
  let builtin_clauses = List.map true_clause b_clauses in *)

  let true_rewrite (l, _, _) =
    Clause (Ocaml_pattern (genr_lhs l), Id "true") in
  let user_clauses =
    List.map true_rewrite (Relation.rewrite fdef.fd_relations) in
  let default = [Clause (underscore, Id "false")] in
  let clauses = user_clauses @ (* builtin_clauses @ *) default in
    is_redex_in cname, args, Match (z, clauses);;

(* Generating the construction function for the (binary) constructor
   given by [fdef].

   We generate the clauses according to the relations recorded to the
   constructor, and add a default insert clause. *)

let genr_constr_function fdef =

  let cname = fdef.fd_uname in

  (* arguments of [construction_function_name cname] *)
  let args = Code.idents_of_exps [z] in

  let insert_fun =
    match inverse fdef.fd_relations with
    | Some (_si, iname, _ename) -> fun x y -> insert_inv cname (constr_fun1 iname x) y
    | None -> insert cname in

  let clauses =
    List.flatten (List.map (genr_fun_2_by_ap fdef) fdef.fd_relations) @
    [Clause (pair x y, insert_fun x y)] in

  construction_function_name cname, args, Match (z, clauses);;

(* Generating the set of recursive definitions that defines the
   construction functions with its appropriate auxilliaries
   (insert_*, delete_*, ...). *)

let genr_fun_2 fdef =

  let cname = fdef.fd_uname in

  let constr_function = genr_constr_function fdef

  and insert_functions = [
    genr_is_redex fdef;
    genr_return cname;
    genr_insert fdef]

  and delete_functions =
    match inverse fdef.fd_relations, neutral fdef.fd_relations with
    | Some (_si, iname, Some ename), _
    | Some (_si, iname, None), Some (_, ename) ->
	[genr_delete cname ename;
	 genr_insert_inv cname iname]
    | _ -> [] in

  constr_function :: delete_functions @ insert_functions;;
